"""Top-level package for rframe."""

__author__ = """Yossi Mosbacher"""
__email__ = 'joe.mosbacher@gmail.com'
__version__ = '0.1.0'


from .rframe import RemoteFrame
from .schema import BaseSchema
from .indexes import *
